{% extends 'Agent/main.jinja2' %}

{% block instructions %}
Determine which tables are necessary to answer the user query, paying special attention to the available columns.
{% if separator %}
Use table names verbatim, and be sure to include the delimiters {{ separator }}, like '<source>{{ separator }}<table>'
{% endif %}
{% endblock %}

{% block context %}
{{ memory["sql_metaset"].max_context }}
{% endblock %}
